#lang racket
;;; sol10.rkt -- avec quelques codes issus du cours 10
;;; Option PF2 en L2-Info, Nice, Automne 2016

;;; utilities
(define-syntax forever
  (syntax-rules ()
    ((forever e ...) (for ([i (in-naturals)]) e ...))))          ; la boucle while True : ...

(define-syntax while
  (syntax-rules ()
    ((while test e ...) (forever #:break (not test) e ...))))    ; fun time :-)

#;(let ((x 0))                                                   ; le #; commente une EXPRESSION
    (while (< x 10) (printf "~a " x) (set! x (add1 x)))
    (printf "\n"))

(define-syntax show
  (syntax-rules ()
    ((show e) (begin (printf "? ~a\n" 'e) (printf "-> ~a\n" e)))))

"--- Le tri de la bulle"

(define (vector-swap! v i j)   ; v[i] <--> v[j]
  (define tmp (vector-ref v i))
  (vector-set! v i (vector-ref v j))
  (vector-set! v j tmp))

(printf "Test de vector-swap! :\n")
(define v (vector 7 5 8 4 3))
(show v)
(show (begin (vector-swap! v 1 3) v))

(define (bubble-sort! v)
  (define n (vector-length v))
  (for ([passe (in-range 1 n)])
    (for ([i (in-range 0 (- n passe))])
      (when (> (vector-ref v i) (vector-ref v (+ i 1)))
        (vector-swap! v i (+ i 1))))
    (printf "--- passe ~a --> v = ~s\n" passe v)))

(printf "\nTest de bubble-sort! sur v = ~a :\n" v)
(bubble-sort! v)
(show v)

"--- Le Triangle de Pascal"

(define (pascal n)    ; on suppose n >= 2
  (define res (build-vector (+ n 1) (lambda (i) (make-vector (+ i 1) 1))))
  (for ([i (in-range 2 (+ n 1))])
    (define v (vector-ref res (- i 1))) 
    (define w (vector-ref res i))
    (for ([j (in-range 1 i)])
      (vector-set! w j (+ (vector-ref v j) (vector-ref v (- j 1))))))
  res)

(show (pascal 6))

(define (pascal-print n)
  (define T (pascal n))
  (for ([i (in-range (+ n 1))])
    (for ([j (in-range (+ i 1))])
      (printf "~a\t" (vector-ref (vector-ref T i) j)))
    (printf "\n")))

(printf "et son affichage :\n")
(pascal-print 6)

"--- Le type abstrait Matrice"

(define (build-matrix nblig nbcol f)   ; M[i,j] = (f i j), i = ligne, j = colonne
  (build-vector nblig (lambda (i) 
                        (build-vector nbcol (lambda (j) (f i j))))))

(define (nb-lignes M)
  (vector-length M)) 

(define (nb-colonnes M)
  (vector-length (vector-ref M 0))) 

(define (matrix-format M)
  (list (vector-length M) (vector-length (vector-ref M 0))))

(define (matrix-ref M i j)     ; i=0..nb-lignes-1 et j=0..nb-colonnes-1
  (vector-ref (vector-ref M i) j)) 

(define (matrix-set! M i j x)
  (vector-set! (vector-ref M i) j x)) 

(define (mat-id n)              ; matrice-unite d'ordre n
  (build-matrix n n (lambda (i j) (if (= i j) 1 0))))

(define (print-matrix M)
  (for ([i (in-range (nb-lignes M))])
    (printf "[")
    (for ([j (in-range (- (nb-colonnes M) 1))])
      (printf "~a\t" (matrix-ref M i j)))
    (printf "~a]\n" (matrix-ref M i (- (nb-colonnes M) 1)))))

(define (mat+ M1 M2)
  (let ((l1 (nb-lignes M1)) (c1 (nb-colonnes M1)) (l2 (nb-lignes M2)) (c2 (nb-colonnes M2)))
    (when (not (and (= l1 l2) (= c1 c2))) (error "Mauvais formats de matrices !"))
    (build-matrix l1 c1 (lambda (i j) (+ (matrix-ref M1 i j) (matrix-ref M2 i j))))))

(define A (build-matrix 3 3 +))
(define B (mat-id 3))
(printf "La matrice A, de format ~a :\n" (matrix-format A))
(print-matrix A)
(printf "La matrice B :\n")
(print-matrix B)
(printf "La matrice A+B :\n")
(print-matrix (mat+ A B))

(define (size-of-gif-image fich)
  (call-with-input-file fich
    (lambda (p-in)
      ; je lis les 6 premiers octets [type de l'image]
      (printf "L'image est de type ~a\n" (for/list ([i (in-range   6)]) (read-char p-in)))
      ; je stocke les 4 octets [char] suivants dans un vecteur [cf cours 9]
      (define v (build-vector 4 (lambda (i) (read-byte p-in))))
      (printf "v = ~a\n" v)
      (printf "L'image a donc pour taille ~a x ~a\n"
              (+ (vector-ref v 0) (* 256 (vector-ref v 1)))
              (+ (vector-ref v 2) (* 256 (vector-ref v 3)))))))

(size-of-gif-image "matrix.gif")    ; 362 x 500

"********************************* TP 11 *********************************"

"--- Exo 1"
;;; La mauvaise programmation de $list->vector provient de l'utilisation de (list-ref L i)
;;; dont la complexite est O(i). Donc le cout de ($list->vector L) est O(n) pour length
;;; suivi d'un O(n**2) pour la boucle for, donc au total O(n**2).
;;; DERNIERE MINUTE : DANS UN BLOG DE JAY MC-CARTHY, JE LIS QUE RACKET OPTIMISE LENGTH MAINTENANT ?
;;; http://jeapostrophe.github.io/2013-07-30-k-from-e-post.html


;;; Dans c) attention a ne pas confondre vector-map qui retourne une copie avec son 
;;; analogue vector-map! qui ferait une mutation du vecteur... 

(define (vector-map_v1 f v)
  (build-vector (vector-length v)
                (lambda (i) (f (vector-ref v i)))))

(printf "Test de vector-map_v1 sur v = ~a:\n" v)
(show (vector-map_v1 add1 v))

(define (vector-map_v2 f v)   ; for/vector <==> for/list mais les resultats sont accumules dans un vecteur
  (for/vector ([x (in-vector v)]) (f x)))    ; une "comprehension de vecteur"...

(printf "Test de vector-map_v2 sur v = ~a:\n" v)
(show (vector-map_v2 add1 v))

;;; Voici un vector-map d'ARITE QUELCONQUE :

(define (vector-map f . LV)    ; LV = liste non vide de vecteurs de meme longueur
  (build-vector (vector-length (car LV))
                (lambda (i) (apply f (map (lambda (v) (vector-ref v i)) LV)))))

(printf "Test de vector-map d'arite quelconque :\n")
(show (vector-map + #(1 2 3) #(4 5 6) #(7 8 9) #(10 11 12)))

"vector-swap est corrige plus haut, dans le tri de la bulle..."
;;; Au contraire de (swap! v1 v2) qui est une MACRO (a cause de l'appel par valeur),
;;; vector-swap! est une FONCTION, car le vecteur v subit une mutation, son adresse
;;; n'etant pas modifiee. La "valeur" d'un vecteur est son adresse, son contenu n'est
;;; obtenu que par affichage du vecteur. Idem en Python ou en C...

"--- Exo 2"

(define (mat* A B)
  (let ((l1 (nb-lignes A)) (c1 (nb-colonnes A)) (l2 (nb-lignes B)) (c2 (nb-colonnes B)))
    (if (not (= c1 l2))
        (error "Produit matriciel impossible : M(~a,~a) * M(~a,~a)" l1 c1 l2 c2)
        (build-matrix l1 c2
                      (lambda (i j)
                        (for/sum ([k (in-range c1)]) 
                          (* (matrix-ref A i k) (matrix-ref B k j))))))))

(define M1 #(#(1 2 3)
             #(5 1 2)))

(define M2 #(#(6 2 0)
             #(1 1 1)
             #(2 0 1)))

(printf "Le produit de la matrice M1 de format (~a,~a) :\n" (nb-lignes M1) (nb-colonnes M1)) 
(print-matrix M1)
(printf "par la matrice M2 de format (~a,~a) :\n" (nb-lignes M2) (nb-colonnes M2))
(print-matrix M2) 
(printf "est egal a la matrice de format (~a,~a) :\n" (nb-lignes M1) (nb-colonnes M2)) 
(print-matrix (mat* M1 M2))
(when  (not (equal? (mat* M1 M2) #(#(14 4 5) #(35 11 3))))   ; comparaison directe de matrices !
  (error "Oups, la fonction mat* est incorrecte !"))

"--- Exo 3 -- Le Drapeau Hollandais de Disjkstra"
;;; https://fr.wikipedia.org/wiki/Algorithme_du_drapeau_tricolore

(define (random-drapeau n)
  (build-vector n (lambda (i) (case (random 3) ((0) 'bleu) ((1) 'blanc) (else 'rouge)))))

(define DRAP (random-drapeau 15))
(printf "Voici un drapeau de longueur ~a:\n~a\n" (vector-length DRAP) DRAP)

(define (cpt-vector-swap! v i j)   ; avec compteur de swap...
  (define tmp (vector-ref v i))
  (vector-set! v i (vector-ref v j))
  (vector-set! v j tmp)
  (set! cpt (+ cpt 1)))

(define (tri-drapeau! v)    ; aucun resultat, mutation de v
  (let ((b 0) (i 0) (r (vector-length v)))
    (while (< i r)
      (case (vector-ref v i)
        ((bleu) (cpt-vector-swap! v b i) (set! b (+ b 1)) (set! i (+ i 1)))
        ((blanc) (set! i (+ i 1)))
        (else (set! r (- r 1)) (cpt-vector-swap! v r i))))))

(define cpt 0)
(tri-drapeau! DRAP)
(printf "et apres le tri (en ~a echanges) :\n~a\n" cpt DRAP)

;(require racket/system)
;(void (system "open http://iml.univ-mrs.fr/~lafont/licence/prog2.pdf"))   ; vous avez open sur votre Unix ?

"--- Exo 4 -- (imperative) fun time, enjoy !"
;;; L'animation est en style imperatif. Le Monde n'intervient pas, le big-bang
;;; n'est la que pour gerer l'horloge et le double-buffer, c'est toujours ca en moins.
;;; Si le monde etait une structure ce serait bien entendu une structure #:mutable,
;;; et le resultat appelerait beaucoup moins le Garbage Collector !

(require 2htdp/image 2htdp/universe rsound)   ; images, animations, sons

(define (anim-drapeau n)
  (let ((SIZE 40)     ; taille d'une case du vecteur
        (BLEU (circle 20 'solid "blue"))
        (BLANC (circle 20 'solid "white"))
        (ROUGE (circle 20 'solid "red"))
        (INIT-SCENE (rectangle (* n 40) 40 'solid "black"))
        (b 0) 
        (i 0) 
        (r n)
        (INIT 0)                   ; le monde est le numero du top d'horloge
        (v (random-drapeau n))     ; code imperatif, mutation sur v
        (ps (make-pstream)))       ; the audience is listening...
    ;;; Le Monde m est bidon, je travaille directement sur ses variables mutables, gasp (mais tres efficace) !
    (define (fini? m)              ; apocalypse now ?
      (>= i r))
    (define (suivant m)            ; mutation du Monde a chaque top d'horloge
      (when (< i r)
        (pstream-play ps (list-ref (list ding crash-cymbal clap-1 snare bassdrum c-hi-hat-1)
                                   (random 6)))
        (case (vector-ref v i)
          ((bleu) (vector-swap! v b i) (set! b (+ b 1)) (set! i (+ i 1)))
          ((blanc) (set! i (+ i 1)))
          (else (set! r (- r 1)) (vector-swap! v r i)))
        (+ m 1)))
    (define (dessiner m)           ; modification de l'image du Monde a chaque top d'horloge
      (define SCENE INIT-SCENE)
      (for ([i (in-range n)])
        (set! SCENE (place-image (case (vector-ref v i)
                                   ((bleu) BLEU) ((blanc) BLANC) (else ROUGE))
                                 (+ 20 (* i 40)) 20 SCENE)))
      SCENE)
    (big-bang INIT    ; le Monde est bidon, c'est en realite l'ensemble des variables MUTABLES
              (name "Drapeau hollandais")
              (on-tick suivant 1/4)
              (on-draw dessiner)
              (stop-when fini?))))

(define LONGUEUR 30)
(define RES (anim-drapeau LONGUEUR))
(printf "Le tri du drapeau de longueur ~a en ~a coups !\n"
        LONGUEUR RES)


